from typing import Type
from pydfs_lineup_optimizer.settings import BaseSettings, LineupPosition
from pydfs_lineup_optimizer.constants import Sport, Site
from pydfs_lineup_optimizer.sites.sites_registry import SitesRegistry
from pydfs_lineup_optimizer.rules import FanduelSingleGameMaxQBRule
from pydfs_lineup_optimizer.sites.fanduel.classic.importer import FanDuelCSVImporter, FanDuelMVPCSVImporter
from pydfs_lineup_optimizer.sites.fanduel.single_game.importer import build_fanduel_single_game_importer


class FanDuelSingleGameSettings(BaseSettings):
    site = Site.FANDUEL_SINGLE_GAME
    budget = 60000
    max_from_one_team = 4
    csv_importer = FanDuelMVPCSVImporter  # type: Type[FanDuelCSVImporter]


@SitesRegistry.register_settings
class FanDuelSingleGameFootballSettings(FanDuelSingleGameSettings):
    sport = Sport.FOOTBALL
    extra_rules = [FanduelSingleGameMaxQBRule]
    positions = [
        LineupPosition('MVP', ('MVP', )),
        LineupPosition('FLEX', ('QB', 'WR', 'RB', 'TE', 'K')),
        LineupPosition('FLEX', ('QB', 'WR', 'RB', 'TE', 'K')),
        LineupPosition('FLEX', ('QB', 'WR', 'RB', 'TE', 'K')),
        LineupPosition('FLEX', ('QB', 'WR', 'RB', 'TE', 'K')),
    ]


@SitesRegistry.register_settings
class FanDuelSingleGameBasketballSettings(FanDuelSingleGameSettings):
    sport = Sport.BASKETBALL
    csv_importer = build_fanduel_single_game_importer(mvp=True, star=True, pro=True)
    positions = [
        LineupPosition('MVP', ('MVP', )),
        LineupPosition('STAR', ('STAR', )),
        LineupPosition('PRO', ('PRO', )),
        LineupPosition('FLEX', ('PG', 'SG', 'SF', 'PF', 'C')),
        LineupPosition('FLEX', ('PG', 'SG', 'SF', 'PF', 'C')),
    ]


@SitesRegistry.register_settings
class FanDuelSingleGameLOLSettings(FanDuelSingleGameSettings):
    sport = Sport.LEAGUE_OF_LEGENDS
    csv_importer = build_fanduel_single_game_importer(mvp=True, star=True, pro=False)
    positions = [
        LineupPosition('MVP', ('MVP', )),
        LineupPosition('STAR', ('STAR', )),
        LineupPosition('UTIL', ('TOP', 'MID', 'ADC', 'JNG', 'SUP')),
        LineupPosition('UTIL', ('TOP', 'MID', 'ADC', 'JNG', 'SUP')),
        LineupPosition('UTIL', ('TOP', 'MID', 'ADC', 'JNG', 'SUP')),
    ]



from typing import Dict, Tuple, List, Iterable, Set, Any, DefaultDict, Optional, TYPE_CHECKING
from collections import Counter, defaultdict
from difflib import SequenceMatcher
from itertools import combinations, chain, permutations
from pydfs_lineup_optimizer.settings import LineupPosition
from pydfs_lineup_optimizer.exceptions import LineupOptimizerException


if TYPE_CHECKING:
    from pydfs_lineup_optimizer.player import Player, LineupPlayer


def list_intersection(first_list: Iterable[Any], second_list: Iterable[Any]) -> bool:
    for el in first_list:
        if el in second_list:
            return True
    return False


def ratio(search_string: str, possible_match: str) -> float:
    search_string = search_string.lower()
    possible_match = possible_match.lower()
    if len(search_string) >= len(possible_match):
        parts = [possible_match]
    else:
        shorter_length = len(search_string)
        num_of_parts = len(possible_match) - shorter_length
        parts = [possible_match[i:i + shorter_length] for i in range(num_of_parts + 1)]
    return max([SequenceMatcher(None, search_string, part).ratio() for part in parts])


def get_positions_for_optimizer(
        positions_list: List[LineupPosition],
        multi_positions_combinations: Optional[Set[Tuple[str, ...]]] = None
) -> Dict[Tuple[str, ...], int]:
    """
    Convert positions list into dict for using in optimizer.
    """
    positions = {}
    positions_counter = Counter([tuple(sorted(p.positions)) for p in positions_list])
    for key in positions_counter.keys():
        min_value = positions_counter[key] + len(list(filter(
            lambda p: len(p.positions) < len(key) and list_intersection(key, p.positions), positions_list
        )))
        positions[key] = min_value
    if not multi_positions_combinations:
        return positions
    #  Create list of required combinations for consistency of multi-positions
    for i in range(2, len(multi_positions_combinations)):
        total_combinations = len(multi_positions_combinations)
        for positions_tuple in combinations(multi_positions_combinations, i):
            flatten_positions = tuple(sorted(set(chain.from_iterable(positions_tuple))))
            multi_positions_combinations.add(flatten_positions)
        if total_combinations == len(multi_positions_combinations):
            break
    multi_positions_combinations.update(positions.keys())
    for i in range(2, len(positions)):
        for positions_tuple in combinations(positions_counter.keys(), i):
            flatten_positions = tuple(sorted(set(chain.from_iterable(positions_tuple))))
            if flatten_positions in positions or flatten_positions not in multi_positions_combinations:
                continue
            min_value = sum(positions[pos] for pos in positions_tuple)
            positions[flatten_positions] = min_value
    return positions


def link_players_with_positions(
        players: List['Player'],
        positions: List[LineupPosition]
) -> Dict['Player', LineupPosition]:
    """
    This method tries to set positions for given players, and raise error if can't.
    """
    positions = positions[:]
    players_with_positions = {}  # type: Dict['Player', LineupPosition]
    players = sorted(players, key=get_player_priority)
    for position in positions:
        players_for_position = [p for p in players if list_intersection(position.positions, p.positions)]
        if len(players_for_position) == 1:
            players_with_positions[players_for_position[0]] = position
            positions.remove(position)
            players.remove(players_for_position[0])
    for players_permutation in permutations(players):
        is_correct = True
        remaining_positions = positions[:]
        for player in players_permutation:
            for position in remaining_positions:
                if list_intersection(player.positions, position.positions):
                    players_with_positions[player] = position
                    remaining_positions.remove(position)
                    break
            else:
                is_correct = False
                break
        if is_correct:
            break
    else:
        raise LineupOptimizerException('Unable to build lineup')
    return players_with_positions


def get_remaining_positions(
        positions: List[LineupPosition],
        unswappable_players: List['LineupPlayer']
) -> List[LineupPosition]:
    """
    Remove unswappable players positions from positions list
    """
    positions = positions[:]
    for player in unswappable_players:
        for position in positions:
            if position.name == player.lineup_position:
                positions.remove(position)
                break
    return positions


def get_players_grouped_by_teams(
        players: Iterable['Player'],
        for_teams: Optional[List[str]] = None,
        for_positions: Optional[List[str]] = None,
) -> DefaultDict[str, List['Player']]:
    players_by_teams = defaultdict(list)  # type: DefaultDict[str, List['Player']]
    for player in players:
        if for_teams is not None and player.team not in for_teams:
            continue
        if for_positions is not None and not list_intersection(player.positions, for_positions):
            continue
        players_by_teams[player.team].append(player)
    return players_by_teams


def process_percents(percent: Optional[float]) -> Optional[float]:
    return percent / 100 if percent and percent > 1 else percent


def get_player_priority(player: 'Player') -> float:
    return float(player.game_info.starts_at.timestamp()) if player.game_info and player.game_info.starts_at else 0.0



from typing import Optional, Dict, List, Type
from pydfs_lineup_optimizer.player import Player
from pydfs_lineup_optimizer.lineup import Lineup
from pydfs_lineup_optimizer.exposure_strategy import BaseExposureStrategy, TotalExposureStrategy


class OptimizationContext:
    def __init__(
            self,
            total_lineups: int,
            players: List[Player],
            max_exposure: Optional[float] = None,
            randomness: bool = False,
            with_injured: bool = False,
            existed_lineups: List[Lineup] = None,
            exposure_strategy: Type[BaseExposureStrategy] = TotalExposureStrategy,
    ):
        self.total_lineups = total_lineups
        self.players = players
        self.remaining_lineups = total_lineups
        self.max_exposure = max_exposure
        self.randomness = randomness
        self.with_injured = with_injured
        self.existed_lineups = existed_lineups or []
        self.lineups = []  # type: List[Lineup]
        self.players_used_fppg = {}  # type: Dict[Player, float]
        self.exposure_strategy = exposure_strategy

    def add_lineup(self, lineup: Lineup) -> None:
        self.lineups.append(lineup)
        self.remaining_lineups -= 1
        self.players_used_fppg = {}



from pydfs_lineup_optimizer.sites.sites_registry import SitesRegistry
from pydfs_lineup_optimizer.sites.draftkings import *  # type: ignore
from pydfs_lineup_optimizer.sites.fanball import *  # type: ignore
from pydfs_lineup_optimizer.sites.fanduel import *  # type: ignore
from pydfs_lineup_optimizer.sites.fantasy_draft import *  # type: ignore
from pydfs_lineup_optimizer.sites.yahoo import *  # type: ignore



from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from pydfs_lineup_optimizer.lineup import Lineup, LineupPlayer


class BaseLineupPrinter:
    def print_lineup(self, lineup: 'Lineup') -> str:
        raise NotImplementedError


class LineupPrinter(BaseLineupPrinter):
    OUTPUT_FORMAT = '{index:>2}. {lineup_position:<5} {name:<30}{positions:<6}{team:<15}{game:<9}' \
                    '{fppg:<15}{salary:<10}\n'

    def _print_game_info(self, player: 'LineupPlayer') -> str:
        game_info = player.game_info
        if game_info:
            return '%s@%s' % (game_info.away_team, game_info.home_team)
        return ''

    def _print_player(self, index: int, player: 'LineupPlayer') -> str:
        return self.OUTPUT_FORMAT.format(
            index=index,
            lineup_position=player.lineup_position,
            name='%s%s' % (player.full_name, '(%s)' % player.roster_order if player.roster_order else ''),
            positions='/'.join(player.original_positions),
            team=player.team,
            game=self._print_game_info(player),
            fppg=round(player.fppg, 3) if player.used_fppg is None else
            '%s(%s)' % (round(player.fppg, 3), round(player.used_fppg, 3)),
            salary=str(player.salary) + '$',
        )

    def _print_footer(self, lineup: 'Lineup') -> str:
        original_projection = lineup.fantasy_points_projection
        actual_projection = lineup.actual_fantasy_points_projection
        footer = 'Fantasy Points %.2f%s\n' % (
            original_projection, '(%.2f)' % actual_projection if actual_projection != original_projection else '')
        if lineup.salary_costs:
            footer += 'Salary %.2f\n' % lineup.salary_costs
        ownerships = [player.projected_ownership for player in lineup if player.projected_ownership]
        if ownerships:
            footer += 'Average Ownership %.1f%%\n' % (sum(ownerships) * 100 / len(ownerships))
        return footer

    def print_lineup(self, lineup):
        res = ''
        for index, player in enumerate(lineup.players, start=1):
            res += self._print_player(index, player)
        res += '\n'
        res += self._print_footer(lineup)
        return res


class IndividualSportLineupPrinter(LineupPrinter):
    OUTPUT_FORMAT = '{index:>2}. {lineup_position:<5} {name:<30}{fppg:<15}{salary:<10}\n'


class DraftKingTiersLineupPrinter(LineupPrinter):
    OUTPUT_FORMAT = '{index:>2}. {lineup_position:<5} {name:<30}{fppg:<15}\n'



from typing import List


class SolverException(Exception):
    pass


class SolverInfeasibleSolutionException(SolverException):
    def __init__(self, invalid_constraints: List[str]):
        self.invalid_constraints = invalid_constraints

    def get_user_defined_constraints(self) -> List[str]:
        return [name for name in self.invalid_constraints if not name.startswith('_')]



from pydfs_lineup_optimizer.settings import BaseSettings, LineupPosition
from pydfs_lineup_optimizer.constants import Sport, Site
from pydfs_lineup_optimizer.sites.sites_registry import SitesRegistry
from pydfs_lineup_optimizer.sites.draftkings.captain_mode.importer import DraftKingsCaptainModeCSVImporter


POSITIONS_WITH_FLEX = [
    LineupPosition('CPT', ('CPT',)),
    LineupPosition('FLEX', ('FLEX',)),
    LineupPosition('FLEX', ('FLEX',)),
    LineupPosition('FLEX', ('FLEX',)),
    LineupPosition('FLEX', ('FLEX',)),
    LineupPosition('FLEX', ('FLEX',)),
]


class DraftKingsCaptainModeSettings(BaseSettings):
    site = Site.DRAFTKINGS_CAPTAIN_MODE
    budget = 50000
    max_from_one_team = 5
    csv_importer = DraftKingsCaptainModeCSVImporter
    positions = [
        LineupPosition('CPT', ('CPT',)),
        LineupPosition('UTIL', ('UTIL',)),
        LineupPosition('UTIL', ('UTIL',)),
        LineupPosition('UTIL', ('UTIL',)),
        LineupPosition('UTIL', ('UTIL',)),
        LineupPosition('UTIL', ('UTIL',)),
    ]


@SitesRegistry.register_settings
class DraftKingsCaptainModeFootballSettings(DraftKingsCaptainModeSettings):
    sport = Sport.FOOTBALL
    positions = POSITIONS_WITH_FLEX[:]


@SitesRegistry.register_settings
class DraftKingsCaptainModeBasketballSettings(DraftKingsCaptainModeSettings):
    sport = Sport.BASKETBALL


@SitesRegistry.register_settings
class DraftKingsLOLSettings(DraftKingsCaptainModeSettings):
    sport = Sport.LEAGUE_OF_LEGENDS
    max_from_one_team = 4
    min_games = 2
    positions = [
        LineupPosition('CPT', ('CPT', )),
        LineupPosition('TOP', ('TOP', )),
        LineupPosition('JNG', ('JNG', )),
        LineupPosition('MID', ('MID', )),
        LineupPosition('ADC', ('ADC', )),
        LineupPosition('SUP', ('SUP', )),
        LineupPosition('TEAM', ('TEAM', )),
    ]


@SitesRegistry.register_settings
class DraftKingsCaptainModeBaseballSettings(DraftKingsCaptainModeSettings):
    sport = Sport.BASEBALL


@SitesRegistry.register_settings
class DraftKingsCaptainModeWNBASettings(DraftKingsCaptainModeSettings):
    sport = Sport.WNBA


@SitesRegistry.register_settings
class DraftLingsCaptainModeSoccerSettings(DraftKingsCaptainModeSettings):
    sport = Sport.SOCCER
    positions = POSITIONS_WITH_FLEX[:]

@SitesRegistry.register_settings
class DraftKingsCaptainModeNHLSettings(DraftKingsCaptainModeSettings):
    sport = Sport.HOCKEY
    positions = POSITIONS_WITH_FLEX[:]




from pydfs_lineup_optimizer.exceptions import LineupOptimizerIncorrectCSV
from pydfs_lineup_optimizer.player import Player
from pydfs_lineup_optimizer.sites.draftkings.classic.importer import DraftKingsCSVImporter


class DraftKingsTiersCSVImporter(DraftKingsCSVImporter):  # pragma: nocover
    def _row_to_player(self, row):
        try:
            name = row['Name'].split()
            player = Player(
                row['ID'],
                name[0],
                name[1] if len(name) > 1 else '',
                row['Roster Position'].split('/'),
                row['TeamAbbrev'],
                0,
                float(row['AvgPointsPerGame']),
                game_info=self._parse_game_info(row),
                original_positions=row['Position'].split('/'),
                **self.get_player_extra(row)
            )
        except KeyError:
            raise LineupOptimizerIncorrectCSV
        return player
